# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import traceback

EX_OK = 0
EX_NORESULT = 1
EX_USAGE = 2
EX_SUBPROCESS = 3
EX_PROCESS = 4
EX_BAD_PLUGIN = 10
EX_BAD_ATTRIBUTE = 11
EX_EXPRESSION = 12

class BaseError(Exception):
    """Base exception for all other exceptions.
    """
    # pylint:disable=redefined-outer-name

    exitcode = None

    def __init__(self, message, traceback=None):
        super().__init__(message)
        self.message = message
        self.traceback = traceback

    @classmethod
    def from_exception(cls, message):
        """Create a BaseError exception with the traceback of the current exception.
        """
        return cls(message, traceback.format_exc())

class NoResultError(BaseError):
    """The --fail option was specified and there were no search results.
    """
    exitcode = EX_NORESULT

class UsageError(BaseError):
    """There was an error in the arguments provided by the user.
    """
    exitcode = EX_USAGE

class SubprocessError(BaseError):
    """One or more --exec or --exec-batch subprocesses had errors.
    """
    exitcode = EX_SUBPROCESS

class ProcessError(BaseError):
    """One or more ff processes had unrecoverable errors.
    """
    exitcode = EX_PROCESS

class BadPluginError(BaseError):
    """A plugin had an unrecoverable error.
    """
    exitcode = EX_BAD_PLUGIN

class BadAttributeError(BaseError):
    """An attribute was specified that is invalid or does not exist.
    """
    exitcode = EX_BAD_ATTRIBUTE

class ExpressionError(BaseError):
    """There was an error in a test definition.
    """
    exitcode = EX_EXPRESSION



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import enum
import binascii

from . import NoData, MissingImport
from .type import *

__all__ = ["NoData", "MissingImport", "Type", "String", "Path", "Number", "FileType", "Mode",
        "Size", "Time", "Duration", "Boolean", "ListOfStrings", "Plugin", "Speed"]


class Speed(enum.Enum):
    """The speed at which a plugin extracts information from an entry. This is used as a hint to
       group faster tests before slower tests in test sequences.
    """
    FAST = 0
    NORMAL = 1
    SLOW = 2


class Plugin:
    """Base class for all plugins.
    """

    author = "Lars Gustäbel"
    email = "lars@gustaebel.de"
    url = "https://github.com/gustaebel/ff/"

    use_cache = False
    cache_tag = None
    speed = Speed.NORMAL
    attributes = {}

    def __init_subclass__(cls):
        super().__init_subclass__()

        cls.name = cls.__name__.lower()

    @classmethod
    def initialize(cls, logger, source, path):
        """Initialize the Plugin class with basic information.
        """
        cls.logger = logger
        cls.source = source
        cls.path = path
        if cls.cache_tag is None:
            cls.cache_tag = cls.get_plugin_cache_tag(path)
        assert isinstance(cls.cache_tag, int) and cls.cache_tag >= 0
        cls.sql_table_name = f"plugin_{cls.name}_{cls.cache_tag}"

    def __init__(self):
        self.setup()

    @classmethod
    def get_plugin_cache_tag(cls, path):
        """Create a checksum of the plugin module file as a tag for the cache database. This way
           the plugin cache is automatically invalidated everytime the module source code changes.
           Return value must be a positive integer.
        """
        with open(path, "rb") as fobj:
            return binascii.crc32(fobj.read())

    def setup(self):
        """Setup the Plugin class. This is called once for every plugin, but only if it will be
           actually used. This is mainly supposed to be used for importing third-party modules that
           may or may not be installed on the system.
        """

    def can_handle(self, entry):
        """Return True if this plugin can handle this particular Entry object. This is supposed to
           reduce calls to process() and minimize the number of needlessly cached entries.
        """
        raise NotImplementedError

    def get_entry_cache_tag(self, entry):
        """Return a value for the Entry object that if it changes causes the cached value to be
           invalidated. The default uses the modification time of the file. The returned value may
           have any type.
        """
        return entry.time

    def cache(self, entry):
        """Extract and prepare data from the Entry object and cache it. It is then passed on to the
           process() method. cache() is only called if use_cache is True. The data returned may
           have any type. Raise a NoData exception if the required data cannot be extracted.
        """
        # pylint:disable=unused-argument
        return None

    def process(self, entry, cached):
        """Generate key-value pairs or return a dict with the values that were "extracted" from the
           entry. The set of keys must match the ones from the attributes dict, but it is not
           necessary that all the keys actually have values. Raise a NoData exception or return an
           empty sequence if the plugin cannot extract the required attributes properly. The cached
           argument is only present when use_cache is True and contains the cached data that was
           prepared in the cache() method. If the cache() method raised a NoData exception
           process() will not be called.
        """
        raise NotImplementedError



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import time
import pickle
import string
import sqlite3

from . import BaseClass


class NOTSET:
    """Return value of the cache if there is no cached value.
    """

class NODATA:
    """Cached value for when there was no value that could have been cached.
    """


class NullCache(BaseClass):
    """A cache that stores nothing and also acts as a base class.
    """

    def close(self):
        """Close the cache.
        """

    def register_plugin(self, plugin_cls):
        """Register a Plugin class in the cache.
        """

    def get(self, plugin, path, tag):
        """Get the object for a specific Plugin class, path with a specific modification time from
           the cache. Return NOTSET if nothing is found.
        """
        # pylint:disable=unused-argument
        return NOTSET

    def set(self, plugin, path, tag, obj):
        """Add an object to the cache for a specific Plugin class, path and modification time.
        """


class Cache(NullCache):
    """A cache that stores pickled Python objects in an sqlite3 database.
    """

    registered = {}

    commit_every_seconds = 30
    commit_every_count = 500

    def __init__(self, context):
        super().__init__(context)

        self.conn = sqlite3.connect(self.context.args.cache, timeout=30)

        self.num_cached_rows = 0
        self.last_commit = time.time()
        self.cached_rows = {}

        self.hits = 0
        self.misses = 0

    def close(self):
        """Commit remaining cached rows and close the database connection.
        """
        if self.num_cached_rows > 0:
            if __debug__:
                self.logger.debug("cache", f"closing cache with {self.num_cached_rows} "\
                        "pending entries")
            self.commit()

        self.conn.close()

        with self.context.cache_hits.get_lock():
            self.context.cache_hits.value += self.hits
        with self.context.cache_misses.get_lock():
            self.context.cache_misses.value += self.misses

    def register_plugin(self, plugin_cls):
        """Register Plugin objects.
        """
        assert plugin_cls.name not in self.registered
        self.init_table(plugin_cls)
        self.registered[plugin_cls.name] = plugin_cls

    def init_table(self, plugin_cls):
        """Check if there is already a table for this plugin in the database. Also drop tables from
           older versions of this plugin.
        """
        # Ask the database if a table exists for this plugin.
        curs = self.conn.execute(
                "select name from sqlite_master where type = 'table' and name like ?",
                (plugin_cls.sql_table_name.rstrip(string.digits) + "%",))

        # Create a table if there is none for this plugin.
        tables = set(row[0] for row in curs)
        if plugin_cls.sql_table_name not in tables:
            for statement in self.get_sql_create_table(plugin_cls):
                self.conn.execute(statement)
            self.conn.commit()
        else:
            tables.remove(plugin_cls.sql_table_name)

        for table in tables:
            if __debug__:
                self.logger.debug("cache", f"Removing old cache table {table!r}")
            self.conn.execute(f"drop table {table}")

    def get_sql_create_table(self, plugin_cls):
        """Return the statements needed to create a table for a specific Plugin class.
        """
        table_name = plugin_cls.sql_table_name
        yield f"create table {table_name} "\
                "(path text not null primary key, tag blob not null, data blob not null)"
        yield f"create index {table_name}_idx on {table_name} (path, tag)"

    def get(self, plugin, path, tag):
        """Return a row of cached values.
        """
        path = path.encode("utf-8", "backslashreplace")
        tag = pickle.dumps(tag, pickle.HIGHEST_PROTOCOL)

        for row in self.conn.execute(
                f"select data from {plugin.sql_table_name} where path = ? and tag = ?",
                (path, tag)):
            self.hits += 1
            return pickle.loads(row[0])

        self.misses += 1
        return super().get(plugin, path, tag)

    def set(self, plugin, path, tag, obj):
        """Write data to the cache.
        """
        path = path.encode("utf-8", "backslashreplace")
        tag = pickle.dumps(tag, pickle.HIGHEST_PROTOCOL)
        data = pickle.dumps(obj, pickle.HIGHEST_PROTOCOL)

        self.cached_rows.setdefault(plugin, []).append((path, tag, data))
        self.num_cached_rows += 1

        if time.time() >= self.last_commit + self.commit_every_seconds or \
                self.num_cached_rows == self.commit_every_count:
            if __debug__:
                self.logger.debug("cache", f"commit cache with {self.num_cached_rows} entries "\
                        f"after {time.time() - self.last_commit:.1f} seconds")
            self.commit()

    def commit(self):
        """We commit cached data in chunks, so that parallel processes will block each other less
           often.
        """
        for plugin, rows in self.cached_rows.items():
            while True:
                try:
                    self.conn.executemany(
                            f"insert or replace into {plugin.sql_table_name} values (?, ?, ?)",
                            rows)
                    self.conn.commit()
                except sqlite3.OperationalError as exc:
                    if "database is locked" in str(exc):
                        self.logger.warning("Database is locked, retrying ...")
                        time.sleep(0.5)
                    else:
                        raise
                else:
                    break

        self.num_cached_rows = 0
        self.cached_rows = {}
        self.last_commit = time.time()

    def clean(self):
        """Remove entries from the database for files that no longer exist.
        """
        for table, in self.conn.execute("select name from sqlite_master where type = 'table'"):
            for path, in self.conn.execute(f"select path from {table}"):
                if not os.path.exists(path):
                    self.conn.execute(f"delete from {table} where path = ?", (path,))

        self.conn.execute("vacuum")
        self.conn.commit()



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import queue
import shutil
import threading
import traceback
import collections
import multiprocessing

from .__version__ import __version__

__copyright__= f"""ff {__version__}
Copyright (C) 2020 Lars Gustäbel <https://github.com/gustaebel/ff>
License GPLv3+: GNU GPL version 3 or later <https://gnu.org/licenses/gpl.html>.

This is free software: you are free to change and redistribute it.
There is NO WARRANTY, to the extent permitted by law."""


TIMEOUT = 0.01
MAX_CPU = multiprocessing.cpu_count()


class NoData(Exception):
    """Exception for when a plugin cannot process an entry.
    """

class MissingImport(Exception):
    """Exception for when a plugin cannot import a third party module.
    """
    def __init__(self, module):
        self.module = module
        super().__init__(module)


OUTPUT_WIDTH = shutil.get_terminal_size((100, 100))[0] - 2


Directory = collections.namedtuple("Directory", "start relpath ignore_paths")
Entries = collections.namedtuple("Entries", "parent entries")


class BaseClass:
    """This class can be used as the base for each class that depends on Context and access to
       args, registry, etc. It offers shortcuts to the each of these components as self.component
       instead of self.context.component without using properties.
    """

    component_names = set(["args", "registry", "excluder", "matcher", "walker"])

    def __init__(self, context):
        self.context = context

    def __getattr__(self, name):
        try:
            attr = getattr(self.context, name)
        except AttributeError as exc:
            raise AttributeError(f"{self.__class__.__name__} object has no attribute {name!r}") \
                    from exc

        if name in self.component_names:
            setattr(self, name, attr)
        return attr



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import pstats
import cProfile

from . import BaseClass, __copyright__
from .walk import Directory, FilesystemWalker
from .cache import Cache, NullCache
from .entry import StartDirectory
from .field import ExecFields, SortFields, CountFields, OutputFields, \
    ExecBatchFields
from .filter import Matcher, Excluder
from .ignore import GitIgnore
from .logger import Logger
from .parser import ParserError
from .context import Context
from .registry import Registry
from .arguments import parse_arguments
from .exceptions import EX_OK, EX_USAGE, EX_PROCESS, EX_SUBPROCESS, \
    BaseError, UsageError, ProcessError, NoResultError, SubprocessError
from .processing import ImmediateExecProcessing, CollectiveExecProcessing, \
    ImmediateConsoleProcessing, CollectiveConsoleProcessing

try:
    import cython
except ImportError:
    pass

class _Base(BaseClass):

    def __init__(self):
        super().__init__(Context())

    def setup_processing(self):
        """Set up the required Processing object.
        """
        raise NotImplementedError

    def setup_context(self, args, error=None, warnings=None):
        """Initialize the central Context object which will be passed to all components.
        """
        self.context.args = args
        self.context.setup()

        self.context.logger = Logger()
        if __debug__:
            self.context.logger.set_debug(args.debug)

        if error is not None:
            self.logger.error(error, EX_USAGE)

        if warnings is not None:
            for warning in warnings:
                self.logger.warning(warning)

    def setup_components(self):
        """Set up all remaining components like the registry, the cache and so on.
        """
        # pylint:disable=too-many-branches

        args = self.context.args

        # Setup the sqlite3 metadata cache.
        if args.cache is None:
            CacheClass = NullCache
        else:
            CacheClass = Cache

        # Once we're on Python >=3.8 we can use context.cache = (cache := CacheClass(context)).
        cache = CacheClass(self.context)
        self.context.cache = cache

        # Register all types and metadata plugins.
        registry = Registry(self.context)
        self.context.registry = registry

        # Replace the special value "file" in -o/--output by a list of all file
        # attributes.
        if "file" in args.output:
            pos = args.output.index("file")
            args.output.pop(pos)
            args.output = args.output[:pos] + [a + ":n" for a in registry.get_file_attributes()] + \
                    args.output[pos:]

        # Turn all options that take fields as arguments in to Fields objects.
        args.output = OutputFields(self.context, args.output)
        if args.sort is not None:
            args.sort = SortFields(self.context, args.sort)
        if args.count is not None:
            args.count = CountFields(self.context, args.count)
        if args.exec is not None:
            args.exec = ExecFields(self.context, args.exec)
        if args.exec_batch is not None:
            args.exec_batch = ExecBatchFields(self.context, args.exec_batch)

        if args.action is not None:
            if args.action == "version":
                print(__copyright__)

            elif args.action == "clean":
                self.context.cache.clean()

            else:
                manpage = getattr(registry, f"get_{args.action}_manpage")()
                manpage.show()

            raise SystemExit(EX_OK)

        elif args.help is not None:
            registry.get_plugin_manpage(args.help).show()
            raise SystemExit(EX_OK)

        try:
            # Setup the matcher from the command line arguments.
            matcher = Matcher(self.context, args.tests)
            self.context.matcher = matcher
        except ParserError as exc:
            raise UsageError(f"unable to parse tests: {exc}") from exc

        # Setup the excluder from the command line arguments.
        excluder = Excluder(self.context, args.exclude)
        self.context.excluder = excluder

    def setup_walker(self):
        """Set up the filesystem walker processes.
        """
        walker = FilesystemWalker(self.context)
        self.context.walker = walker

        if __debug__:
            self.logger.debug("info", "Directories to search:")

        # Preload the in-queue with the path arguments.
        for path in self.context.args.directories:
            if self.context.args.absolute_path:
                path = os.path.abspath(path)

            if self.context.args.no_parent_ignore:
                # Don't use ignore files from parent directories.
                ignores = []
            else:
                # Find ignore files in parent directories.
                if self.context.args.ignore:
                    ignores = GitIgnore.from_parent_directories(self.context.args.ignore_files,
                            os.path.dirname(path))
                else:
                    ignores = []

            if __debug__:
                self.logger.debug("info",
                        f"  {path if path != '.' else os.path.abspath(path)}")
                for ignore in ignores:
                    self.logger.debug("info",
                            f"    found ignorefile in parent: {ignore}")

            walker.put([Directory(StartDirectory(self.context.args, path), "", ignores)])

    def setup(self, args, error=None, warnings=None):
        """Set up context, components, processing and walkers.
        """
        self.setup_context(args, error, warnings)
        self.setup_components()
        self.setup_processing()
        if __debug__:
            self.show_debug_info()
        self.setup_walker()

    def show_debug_info(self):
        """Show general debug information and the tests to be performed.
        """
        self.logger.debug("info", f"Executable: {sys.argv[0]}.")

        try:
            if cython.compiled:
                self.logger.debug("info", "This is a cythonized build.")
        except NameError:
            pass

        self.logger.debug("info", f"Using {self.context.processing.__class__.__name__}.")

        if not self.context.excluder.is_empty():
            self.logger.debug("info", "Exclude Sequence:")
            for line in self.context.excluder.parser.format():
                self.logger.debug("info", "  " + line)

        if not self.context.matcher.is_empty():
            self.logger.debug("info", "Test Sequence:")
            for line in self.context.matcher.parser.format():
                self.logger.debug("info", "  " + line)


class Main(_Base):
    """The main entry point for the ff(1) script.
    """

    def __init__(self):
        super().__init__()

        try:
            self.setup(*parse_arguments())
        except BaseError as exc:
            self.handle_exception(exc)

    def loop(self):
        """Do all the processing and exit.
        """
        try:
            self.walk()
        except BaseError as exc:
            self.handle_exception(exc)

    def handle_exception(self, exc):
        """Print a BaseError exception using the correct formatting and exit.
        """
        if exc.traceback is not None:
            self.logger.exception(exc.message, exc.traceback, exc.exitcode)
        elif isinstance(exc, NoResultError):
            raise SystemExit(exc.exitcode)
        else:
            self.logger.error(exc.message, exc.exitcode)

    def setup_processing(self):
        args = self.context.args

        # Set up processing. There are two modes of operation: Either collect all results and
        # process them in their entirety, or process results immediately as they come in.
        if args.sort or args.count or args.limit or args.exec_batch or args.json == "json":
            if args.exec or args.exec_batch:
                self.context.processing = CollectiveExecProcessing(self.context)
            else:
                self.context.processing = CollectiveConsoleProcessing(self.context)
        elif args.exec:
            self.context.processing = ImmediateExecProcessing(self.context)
        else:
            self.context.processing = ImmediateConsoleProcessing(self.context)

    def walk(self):
        """Walk through the filesystem and process the results.
        """
        # pylint:disable=too-many-branches

        try:
            if __debug__ and self.context.args.profile:
                # Start profiling. No multiprocessing is involved, Walker.loop() is run in the main
                # thread.
                profiler = cProfile.Profile()
                profiler.enable()
                self.context.walker.loop(0)
            else:
                self.context.walker.start_processes()

            # Collect Entry objects from the Walker.
            self.context.processing.loop()

            # Finalize and put out the result if --sort or --exec-batch is involved.
            self.context.processing.finalize()

            if __debug__ and self.context.args.profile:
                # Stop profiling and print the statistics.
                profiler.disable()
                stats = pstats.Stats(profiler, stream=sys.stderr)
                stats.sort_stats("cumulative")
                stats.print_stats(.1)

        except KeyboardInterrupt as exc:
            # Stop all processes immediately.
            self.context.stop()
            if __debug__:
                raise
            else:
                raise SystemExit("keyboard interrupt") from exc

        except BrokenPipeError:
            self.context.stop()

        self.context.close()
        self.context.walker.close()
        self.context.processing.close()

        if __debug__:
            hits = self.context.cache_hits.value
            misses = self.context.cache_misses.value
            if hits or misses:
                self.logger.debug("cache", f"Cache stats: {hits} hits, {misses} misses")
            else:
                self.logger.debug("cache", "Cache was not used")

        if self.context.exitcode == EX_SUBPROCESS:
            if self.args.exec:
                succeeded = self.context.process_succeeded.value
                failed = self.context.process_failed.value
                raise SubprocessError("One or more -x/--exec commands had errors "\
                        f"(success: {succeeded}, failure: {failed})")
            else:
                raise SubprocessError("The -X/--exec-batch command had an error")

        elif self.context.exitcode == EX_PROCESS:
            raise ProcessError("One or more ff processes had unrecoverable errors! "\
                    "Result is probably incomplete!")

        elif self.args.fail and not self.context.has_result():
            raise NoResultError("there were no search results")



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

# pylint:disable=redefined-builtin

import re
import sys
import textwrap

from . import MAX_CPU, OUTPUT_WIDTH
from .exceptions import EX_USAGE


class ArgumentError(Exception):
    """Exception raised if the ArgumentParser is misconfigured or argument parsing fails.
    """


def type_number(string):
    """Parse a number.
    """
    return int(string)


def type_positive_number(string):
    """Parse a number greater than zero.
    """
    num = type_number(string)
    if num < 0:
        raise ValueError("number must be greater than zero")
    return num


def type_jobs(string):
    """Parse the number of jobs using a default of MAX_CPU if it is zero.
    """
    num = type_positive_number(string)
    if num == 0:
        num = MAX_CPU
    return max(min(num, MAX_CPU), 1)


def type_list(string):
    """Parse a comma separated string into a list.
    """
    return [s.strip() for s in string.split(",")]


def type_set(string):
    """Parse a comma separated string into a set.
    """
    return {s.strip() for s in string.split(",")}


regex_range = re.compile(r"^(?:(?P<single>\d+)|(?P<start>\d*)-(?P<stop>\d*))$")

def type_ranges(string):
    """Parse a comma separated list of ranges.
    """
    segments = []
    for range_ in type_list(string):
        match = regex_range.match(range_)
        if match is None:
            raise ValueError("invalid range")

        if match.group("single") is not None:
            number = int(match.group("single"))
            segments.append((number, number))

        else:
            start = match.group("start")
            if start:
                start = int(start)
            else:
                start = 0

            stop = match.group("stop")
            if stop:
                stop = int(stop)
                if stop < start:
                    stop, start = start, stop
            else:
                stop = None

            segments.append((start, stop))

    return segments


regex_slice = re.compile(r"^(?:(?P<start>\d*|-?\d+):(?P<stop>\d*|-?\d+)"\
                         r"|(?P<pagesize>\d+),(?P<page>\d+))$")

def type_slice(string):
    """Parse python-like slice and pagesize,page notation.
    """
    match = regex_slice.match(string)
    if match is None:
        raise ArgumentError(f"invalid slice {string!r}")

    if match.group("start") is not None:
        start = match.group("start")
        if start:
            start = int(start)
        else:
            start = 0

        stop = match.group("stop")
        if stop:
            stop = int(stop)
        else:
            stop = None

    else:
        pagesize = int(match.group("pagesize"))
        page = int(match.group("page"))
        start = pagesize * page
        stop = pagesize * (page + 1)

    return start, stop


class Option:
    """An option container similar to the one in argparse.
    """
    # pylint:disable=too-many-instance-attributes,too-many-arguments

    def __init__(self, strings, action, help, const, choices, type, dest, default, metavar):
        if action == "arguments":
            if len(strings) > 1 or strings[0].startswith("-"):
                raise ArgumentError("misconfigured 'arguments' option")
            self.strings = strings
        else:
            self.strings = tuple(self.check_strings(strings))

        self.action = action
        self.help = help
        self.const = const
        self.choices = choices
        self.type = type
        self.default = default

        if dest is not None:
            self.dest = dest
        else:
            self.dest = self.get_name()

        if self.action in ("store", "store_optional", "append", "arguments"):
            if metavar is not None:
                self.metavar = metavar
            else:
                self.metavar = f"<{self.get_name()}>"
        else:
            self.metavar = None

    def get_name(self):
        """Get the default name for this option.
        """
        for string in self.strings:
            if not string.startswith("-") or string.startswith("--"):
                return string.lstrip("-").replace("-", "_")
        raise ArgumentError(f"missing long option {self.repr}")

    def check_strings(self, strings):
        """Check the validity of the option strings.
        """
        for string in strings:
            if not string:
                raise ArgumentError("empty option string")
            elif string.startswith("--"):
                yield string
            elif string.startswith("-"):
                if len(string) > 2:
                    raise ArgumentError(f"malformed short option {string!r}")
                yield string
            else:
                raise ArgumentError(f"invalid option {string!r}")

    @property
    def repr(self):
        """Return a representation of the option.
        """
        return "/".join(self.strings)


class Namespace(dict):
    """Container for the values assembled from the command line arguments.
    """

    def set(self, key, value):
        """Set the value for key.
        """
        self[key] = value

    def append(self, key, value):
        """Append the value to key.
        """
        container = self[key]
        if container is None:
            self[key] = []
        self[key].append(value)

    def __getattr__(self, key):
        return self[key]

    def __repr__(self):
        return f"Namespace({', '.join(f'{key}={value!r}' for key, value in sorted(self.items()))})"


class ArgumentParser:
    """Parse command line arguments in way similar to argparse.ArgumentParser.
    """

    usage = "ff [<options>] [<test/directory> ... | -D <directory> ...]"

    def __init__(self):
        self.help_output = []
        self.options = []
        self.namespace = Namespace()

    def error(self, message):
        """Print an error message to standard error output and exit with EX_USAGE.
        """
        print(f"usage: {self.usage}", file=sys.stderr)
        print(f"ff: error: {message}", file=sys.stderr)
        sys.exit(EX_USAGE)

    def print_usage(self, file=sys.stdout):
        """Print the usage to standard output.
        """
        print(self.usage, file=file)

    def print_help(self, file=sys.stdout):
        """Print the help to standard output.
        """
        print(f"usage: {self.usage}", file=file)

        for item in self.help_output:
            if isinstance(item, Option):
                option_string = self.format_option(item)

                help_text = textwrap.fill(item.help, width=OUTPUT_WIDTH,
                        initial_indent=" " * 27, subsequent_indent=" " * 27)

                if len(option_string) > 25:
                    print(f"  {option_string}", file=file)
                    print(help_text, file=file)
                else:
                    print(f"  {option_string:25s}{help_text.lstrip()}", file=file)
            else:
                print(file=file)
                print(f"{item}:", file=file)

    def format_option(self, option):
        """Format an Option object.
        """
        if option.metavar is not None:
            if option.action == "arguments":
                return f"{option.metavar}"
            elif option.action == "store_optional":
                return f"{', '.join(option.strings)} [{option.metavar}]"
            else:
                return f"{', '.join(option.strings)} {option.metavar}"
        else:
            return f"{', '.join(option.strings)}"

    def print_manpage_help(self, file):
        """Produce help output suitable for inclusion in the manual page.
        """
        for item in self.help_output:
            if isinstance(item, Option):
                option_string = self.format_option(item)
                print(f"    {option_string}  {item.help}", file=file)
            else:
                print(file=file)
                print(f"    {item}:", file=file)

    def add_group(self, group_name):
        """Add a group separator the help output.
        """
        self.help_output.append(group_name)

    def add_option(self, *strings, action="store", help, const=None, choices=None, type=str,
            dest=None, default=None, metavar=None):
        """Add an option to the ArgumentParser.
        """
        # Check if all the option strings are unique.
        for string in strings:
            for option in self.options:
                if string in option.strings:
                    raise ArgumentError(f"option {string!r} already exists")

        # The "arguments" action must be specified only once.
        if action == "arguments":
            if any(o.action == "arguments" for o in self.options):
                raise ArgumentError("no support for more than one 'arguments' action")

        # Implicitly use an empty list for certain actions.
        if action in ("append", "arguments") and default is None:
            default = []

        # Create the Option object and add it to the list of options and the help output.
        option = Option(strings, action, help, const, choices, type, dest, default, metavar)
        self.options.append(option)
        self.help_output.append(option)

    def initialize(self):
        """Prepare the namespace with default values.
        """
        for option in self.options:
            if option.dest not in self.namespace:
                self.namespace.set(option.dest, option.default)

    def get_option(self, string):
        """Return an Option object for a specific string.
        """
        for option in self.options:
            if option.action == "arguments":
                continue
            elif string in option.strings:
                return option

        raise ArgumentError(f"unknown option {string!r}")

    def get_arguments_option(self):
        """Return the Option object that contains the positional arguments.
        """
        for option in self.options:
            if option.action == "arguments":
                return option

        raise ArgumentError("argument found but parser takes no arguments")

    def parse_combined_short_options(self, argv):
        """Disassemble a bundle of short options.
        """
        arg = argv.pop(0)[1:]
        i = 0
        while arg:
            short_option = f"-{arg[0]}"
            option = self.get_option(short_option)

            if option.action in ("store", "store_optional", "append"):
                argv.insert(i, short_option)
                argv.insert(i + 1, arg[1:])
                break
            else:
                argv.insert(i, short_option)

            arg = arg[1:]
            i += 1
        return argv

    def parse_long_option_with_value(self, argv):
        """Parse a long option with an attached argument.
        """
        arg = argv.pop(0)

        long_option, value = arg.split("=", 1)
        option = self.get_option(long_option)

        if option.action in ("store", "store_optional", "append"):
            argv.insert(0, long_option)
            argv.insert(1, value)
        else:
            raise ArgumentError(f"option {long_option} does not take an argument")

        return argv

    def parse_args(self, argv):
        """Parse a list of command line arguments.
        """
        argv = argv.copy()

        self.initialize()

        try:
            while argv:
                arg = argv[0]

                if arg.startswith("-"):
                    if arg[:2] != "--" and len(arg) > 2:
                        # This is not a long option but a bundle of short options.
                        argv = self.parse_combined_short_options(argv)

                    elif arg[:2] == "--" and "=" in arg:
                        # This is a long option with an attached argument.
                        argv = self.parse_long_option_with_value(argv)

                    self.parse_option(argv)

                else:
                    self.parse_arguments(argv)

        except ArgumentError as exc:
            self.error(exc)

        return self.namespace

    def get_argument(self, option, argv):
        """Fetch the next argument from the argument list.
        """
        if not argv:
            raise ArgumentError(f"{option.repr} requires an argument")
        else:
            return option.type(argv.pop(0))

    def get_optional_argument(self, option, argv):
        """Fetch an optional argument from the argument list.
        """
        if not argv or argv[0].startswith("-"):
            raise ArgumentError(f"{option.repr} requires an argument")
        else:
            return option.type(argv.pop(0))

    def parse_option(self, argv):
        """Parse a single option and act according to its action.
        """
        option = self.get_option(argv.pop(0))

        if option.action == "store_true":
            self.namespace.set(option.dest, True)

        elif option.action == "store_const":
            self.namespace.set(option.dest, option.const)

        elif option.action == "append":
            self.namespace.append(option.dest, self.get_argument(option, argv))

        elif option.action == "store":
            self.namespace.set(option.dest, self.get_argument(option, argv))

        elif option.action == "store_optional":
            if argv and not argv[0].startswith("-"):
                self.namespace.set(option.dest, self.get_optional_argument(option, argv))
            else:
                self.namespace.set(option.dest, option.type(option.const))

        elif option.action == "store_remainder":
            if not argv:
                raise ArgumentError(f"{option.repr} requires an argument")

            while argv:
                self.namespace.append(option.dest, argv.pop(0))

        else:
            raise ArgumentError(f"unknown action {option.action!r}")

    def parse_arguments(self, argv):
        """Parse the positional arguments.
        """
        option = self.get_arguments_option()

        if self.namespace.get(option.dest):
            raise ArgumentError("intermixing arguments with options is not supported")

        while argv and not argv[0].startswith("-"):
            self.namespace.append(option.dest, argv.pop(0))



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import shlex

from . import MAX_CPU
from .ignore import IGNORE_FILES
from .argparse import ArgumentParser, type_set, type_jobs, type_list, \
    type_slice, type_ranges
from .exceptions import EX_OK, UsageError


class Defaults:
    """Provide the defaults for both the ArgumentParser as well as the API.
    """

    case = "smart"
    case_choices = ["smart", "ignore", "sensitive"]
    follow_symlinks = False
    jobs = MAX_CPU
    cache = os.path.expanduser("~/.cache/ff.db")
    si = False


def create_parser():
    """Create the ArgumentParser object.
    """
    # pylint:disable=too-many-statements
    parser = ArgumentParser()

    parser.add_group("Global options")
    parser.add_option("tests", action="arguments", metavar="<test/directory>",
            help="A test for file matching or a the path to a directory to search.")
    if __debug__:
        parser.add_option("--profile", action="store_true", default=False,
                help="Do a profiling run on the given arguments and suppress the output.")
        parser.add_option("--debug", type=type_list, default=["none"], metavar="<categories>",
                help="Show debug messages. Specify either 'all', 'none' or a comma-separated list "\
                     "of <categories>, default is to show 'none'.")
    parser.add_option("--cache", default=Defaults.cache, metavar="<path>",
            help="Location of the metadata cache, the default is ~/.cache/ff.db.")
    parser.add_option("--no-cache", action="store_const", dest="cache", const=None,
            help="Do not use the metadata cache.")
    parser.add_option("--clean-cache", action="store_const", const="clean", dest="action",
            help="Remove stale entries from the metadata cache.")
    parser.add_option("-j", "--jobs", type=type_jobs, default=Defaults.jobs, metavar="<num>",
            help="Set number of processes to use for searching and executing "\
                 "(default: the number of CPU cores).")
    parser.add_option("-D", "--directory", action="append", metavar="<path>", dest="directories",
            default=[], help="Search entries in this path (default is current directory). May be "\
                             "specified multiple times.")

    parser.add_group("Commands")
    parser.add_option("-h", "--help", action="store_optional", const="all", metavar="<plugin>",
            help="Show this help message or the help message for a particular plugin.")
    parser.add_option("--version", action="store_const", const="version", dest="action",
            help="Show program's version number and exit.")
    parser.add_option("--help-full", action="store_const", const="full", dest="action",
            help="Show a full help in man page format.")
    parser.add_option("--help-attributes", action="store_const", const="attributes",
            dest="action",
            help="Show a list of available attributes to use for searching, sorting and output.")
    parser.add_option("--help-plugins", action="store_const", const="plugins", dest="action",
            help="Show a list of available plugins.")
    parser.add_option("--help-types", action="store_const", const="types", dest="action",
            help="Show a list of available types.")

    parser.add_group("Search options")
    parser.add_option("-e", "--exclude", action="append", default=[], metavar="<test>",
            help="Exclude entries that match the given test.")
    parser.add_option("-H", "--hide", action="store_true", default=False,
            help="Do not show hidden files and directories.")
    parser.add_option("-I", "--ignore", action="store_true", default=False,
            help="Do not show files that are excluded by patterns from .{,git,fd,ff}ignore files.")
    parser.add_option("--no-parent-ignore", action="store_true", default=False,
            help="Do not read patterns from ignore files from parent directories.")
    parser.add_option("--ignore-files", type=type_set, metavar="<filename-list>",
            default=IGNORE_FILES,
            help="A comma-separated list of filenames that contain ignore patterns. Default is "\
                f"{','.join(sorted(IGNORE_FILES))}.")
    parser.add_option("-d", "--depth", type=type_ranges, metavar="<range>",
            help="Show only files that are located at a certain depth level of the directory "\
                 "tree that is within the given <range>. A <range> is a string of the form "\
                 "'<start>-<stop>'. <start> and <stop> are optional and may be omitted. "\
                 "<range> may also be a single number. It is possible to specify multiple "\
                 "ranges separated by comma.")
    parser.add_option("-c", "--case", choices=Defaults.case_choices, dest="case",
            default=Defaults.case, metavar="<mode>",
            help="How to treat the case of text attributes (smart, ignore or sensitive).")
    parser.add_option("-L", "--follow", action="store_true", dest="follow_symlinks",
            default=Defaults.follow_symlinks, help="Follow symbolic links.")
    parser.add_option("--one-file-system", "--mount", "--xdev", action="store_true",
            default=False,
            help="Do not descend into different file systems.")

    parser.add_group("Output options")
    parser.add_option("--fail", action="store_true", default=False,
            help="Fail with exitcode 1 if there are no search results.")
    parser.add_option("-x", "--exec", action="store_remainder", metavar="<cmd>",
            help="Execute a command for each search result.")
    parser.add_option("-X", "--exec-batch", action="store_remainder", metavar="<cmd>",
            help="Execute a command with all search results at once.")
    parser.add_option("--halt", "--halt-on-error", choices=["never", "soon", "now"],
            default="never", metavar="<when>",
            help="When to terminate in case one of the -x/--exec commands fails: "\
                 "*never*, soon, now.")
    parser.add_option("-C", "--color", choices=["auto", "never", "always"],
            default="never" if "NO_COLOR" in os.environ else "auto",
            metavar="<when>", help="When to use colors: never, *auto*, always.")
    parser.add_option("-a", "--absolute-path", action="store_true", default=False,
            help="Show absolute instead of relative paths.")
    parser.add_option("-0", "--print0", action="store_const", const="\0", dest="newline",
            default="\n", help="Separate results by the null character.")
    parser.add_option("-v", "--verbose", action="store_const", dest="output", default=["path"],
            const=["mode:h", "links", "user:h", "group:h", "size:5h", "time:h", "path:h"],
            help="Produce output similar to `ls -l`.")
    parser.add_option("-S", "--sort", action="store_optional", type=type_list, const="file.path",
            metavar="<attribute-list>", help="Sort entries by path or any other attribute.")
    parser.add_option("-R", "--reverse", action="store_true", default=False,
            help="Reverse the sort order.")
    parser.add_option("--count", action="store_optional", type=type_list,
            const="", metavar="<attribute-list>",
            help="Count the attributes from <attribute-list> and print statistics, "\
                 "instead of the result, the default is to count the total size and "\
                 "the file types of the entries found. Add --json for JSON output.")
    parser.add_option("-l", "--limit", action="store", type=type_slice, metavar="<slice>",
            help="Limit output to a slice of the search results. A <slice> is defined by either "\
                 "'<start>:<stop>' or '<pagesize>,<page>'. In the first form both <start> and "\
                 "<stop> may be omitted which means they default to the beginning and end of the "\
                 "list of results respectively. If <start> or <stop> are negative their position "\
                 "is relative to the end of the list. The second form creates a slice of "\
                 "<pagesize> results starting at a page offset specified by <page> (starting with "\
                 "0). Note: It is recommended to sort the result list with -S/--sort when using "\
                 "this option, otherwise the order of results between consecutive runs is "\
                 "probably inconsistent.")
    parser.add_option("-1", action="store_const", const=(0, 1), dest="limit",
            help="Limit output to only the first entry.")
    parser.add_option("-o", "--output", type=type_list, metavar="<attribute-list>",
            help="Print each entry by using a template of comma-separated attributes. "\
                 "The special value 'file' stands for all file attributes.")
    parser.add_option("--sep", default=" ", metavar="<string>", dest="separator",
            help="Separate each attribute of --output with <string>, default is a single space.")
    parser.add_option("--all", action="store_true", default=False,
            help="Show all entries including the ones with missing attribute values.")
    parser.add_option("--json", action="store_const", const="json", dest="json",
            help="Print attributes as one big json object to stdout.")
    parser.add_option("--jsonl", "--ndjson", action="store_const", const="jsonl", dest="json",
            help="Print attributes as jsonl (one json object per line) to stdout.")
    parser.add_option("--si", action="store_true", default=Defaults.si,
            help="Parse and print file sizes in units of 1K=1000 bytes instead of 1K=1024 bytes.")

    return parser


def collect_arguments():
    """Join arguments for the ArgumentParser from the FF_OPTIONS environment variable with the ones
       from the command line.
    """
    ff_options = os.environ.get("FF_OPTIONS")
    if ff_options:
        argv = shlex.split(ff_options)
    else:
        argv = []
    argv += sys.argv[1:]
    return argv


class ArgumentsPostProcessor:
    """Check and postprocess the arguments in an argparse.Namespace.
    """

    def __init__(self, args):
        self.args = args

    def process(self):
        """Check and postprocess the arguments.
        """
        self.process_directories()
        self.process_arguments()
        self.check_for_errors()
        return self.collect_warnings()

    def check_for_errors(self):
        """Check some of the arguments for semantic conflicts with other arguments.
        """
        if __debug__:
            if self.args.profile and (self.args.exec or self.args.exec_batch):
                raise UsageError("You cannot use --exec or --exec-batch together with --profile!")

        if self.args.count and (self.args.exec or self.args.exec_batch):
            raise UsageError("You cannot use --exec or --exec-batch together with --count!")

        if self.args.count and self.args.limit is not None:
            raise UsageError("You cannot use --limit together with --count!")

    def collect_warnings(self):
        """Check some of the arguments for semantic conflicts with other arguments.
        """
        warnings = []

        if self.args.output != ["path"] and (self.args.exec or self.args.exec_batch):
            self.args.output = ["path"]
            warnings.append(
                    "Switching off --output, it has no effect with --exec and --exec-batch.")

        if self.args.count:
            if self.args.sort:
                self.args.sort = None
                warnings.append("Switching off --sort, it has no effect with --count.")

            if self.args.output != ["path"]:
                self.args.output = ["path"]
                warnings.append("Switching off --output, it has no effect with --count.")

        if self.args.limit is not None and not self.args.sort:
            warnings.append(
                    "It is recommended to use --limit together with --sort to ensure a stable "\
                    "order of entries.")

        return warnings

    def process_directories(self):
        """Arrange directory arguments and check them for validity.
        """
        if not self.args.directories:
            # Check which arguments are existing directories and append them to
            # self.args.directories. We allow directory arguments only at the start or the end of
            # the list of tests.
            for tests in (list(self.args.tests), reversed(self.args.tests)):
                for test in tests:
                    if os.sep in test and os.path.isdir(test):
                        self.args.tests.remove(test)
                        directory = os.path.normpath(test)
                        self.args.directories.append(directory)
                    else:
                        break

        # Default to the current directory if no directory arguments are specified or detected.
        if not self.args.directories:
            self.args.directories = ["."]

        # Check if directory arguments are sub-directories of one another.
        for directory in sorted(self.args.directories, reverse=True):
            for subdir in self.args.directories:
                if subdir == directory:
                    continue
                # XXX With --one-file-system it would be okay to specify a directory argument that
                # only looks like a subdirectory of another argument but actually lives on a
                # different mount. We should take this into account here.
                abs_directory = os.path.abspath(directory)
                abs_subdir = os.path.abspath(subdir)
                if os.path.commonpath([abs_directory, abs_subdir]) == abs_subdir:
                    raise UsageError(f"{directory!r} is a sub-directory of {subdir!r}")

    def process_depth_argument(self):
        """Translate the argument to -d/--depth to a test expression.
        """
        # Look for an upper limit in the given ranges and exclude everything above it, so we go
        # only as deep into the tree as necessary.
        if not [stop for start, stop in self.args.depth if stop is None]:
            stop = max(stop for start, stop in self.args.depth)
            self.args.exclude.append(f"depth+{stop}")

        # Construct a set of tests for all the given ranges and embed the existing set of tests
        # in it.
        tests = []
        tests.append("{{")

        for i, (start, stop) in enumerate(self.args.depth):
            if start == stop:
                tests.append(f"depth={start}")
            else:
                tests.append(f"depth+={start}")
                if stop is not None:
                    tests.append(f"depth-={stop}")

            if i < len(self.args.depth) - 1:
                tests.append("OR")

        tests.append("}}")

        if self.args.tests:
            tests.append("{{")
            tests += self.args.tests
            tests.append("}}")

        self.args.tests = tests

    def process_arguments(self):
        """Check existing arguments and arrange them in specific ways.
        """
        if self.args.json:
            self.args.color = "never"
            self.args.absolute_path = True

        if self.args.ignore:
            self.args.exclude.append("ignored=yes")

        if self.args.one_file_system:
            self.args.exclude.append("samedev=no")

        if self.args.hide:
            self.args.exclude.append("hide=yes")

        if self.args.depth:
            self.process_depth_argument()

        if self.args.count == [""]:
            if self.args.json is not None:
                self.args.count = ["file.size", "file.type"]
            else:
                self.args.count = ["file.size:h", "file.type"]


def parse_arguments():
    """Parse the arguments from the command line, check for conflicts and postprocess them for
       later use.
    """
    parser = create_parser()

    argv = collect_arguments()
    args = parser.parse_args(argv)

    if args.help == "all":
        # Help on plugins is taken care of in libff/find.py.
        parser.print_help()
        raise SystemExit(EX_OK)

    processor = ArgumentsPostProcessor(args)

    error = None
    warnings = []
    try:
        warnings = processor.process()
    except UsageError as exc:
        error = exc.message

    return args, error, warnings



# -----------------------------------------------------------------------
#
# ff - a tool for finding files in the filesystem
# Copyright (C) 2020 Lars Gustäbel <lars@gustaebel.de>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
#
# -----------------------------------------------------------------------

import os
import sys
import json
import collections

from . import BaseClass

Field = collections.namedtuple("Field", "attribute width modifier")
String = collections.namedtuple("String", "value length")


class BaseConsole(BaseClass):
    """The base class for all Console classes.
    """

    def __init__(self, context):
        super().__init__(context)

        self.fields = self.args.output

    def process(self, entry):
        """Create a record from the Entry object by fetching all required attribute values and
           print the record to standard output.
        """
        raise NotImplementedError

    def close(self):
        """Do some final operations.
        """


class NullConsole(BaseConsole):
    """Do not format and print anything.
    """

    def process(self, entry):
        pass


class JsonlConsole(BaseConsole):
    """Print entries as newline-delimited JSON objects.
    """

    def __init__(self, context):
        super().__init__(context)

        if __debug__ and self.args.profile:
            # Suppress output when profiling but still go through all the moves.
            self.print_record = lambda record: None
        else:
            self.print_record = self._print_record

    def process(self, entry):
        """Print a JSON record for the Entry object.
        """
        self.print_record(self.fields.to_dict(entry))

    def _print_record(self, record):
        """Dump a newline delimited JSON object to standard output.
        """
        json.dump(record, sys.stdout)
        print()
        sys.stdout.flush()


class JsonConsole(JsonlConsole):
    """Print all entries as one big JSON object.
    """

    def __init__(self, context):
        super().__init__(context)

        print("[", end="")
        self.print_comma = False

    def _print_record(self, record):
        """Dump a newline delimited JSON object to standard output.
        """
        if self.print_comma:
            print(",", end="")
        json.dump(record, sys.stdout)
        self.print_comma = True

    def close(self):
        print("]")


class Console(BaseConsole):
    """Print entries as separator delimited records of fields.
    """

    encoding = sys.getfilesystemencoding()

    def __init__(self, context):
        super().__init__(context)

        self.field_separator = self.args.separator.replace("\\t", "\t").replace("\\n", "\n")

        if __debug__ and self.args.profile:
            self.write_line = lambda line: None
        else:
            self.write_line = self._write_line

    def render_field(self, entry, field):
        """Render a field for output, i.e. get the value from the plugin and convert it to an
           individual form using its Type. If required turn it into a human-readable form.
        """
        if field.attribute == ("file", "dir"):
            return self.render_directory(entry.dir)

        elif field.attribute == ("file", "name"):
            return self.render_name(entry)

        elif field.attribute == ("file", "path"):
            if field.modifier == "h":
                return self.render_path(entry, with_link=True)
            else:
                return self.render_path(entry)

        else:
            value, type_cls = self.registry.get_attribute_and_type(entry, field.attribute)
            string = type_cls.output(self.args, field.modifier, value)

            return String(string, len(string))

    def render_path(self, entry, with_link=False):
        """Render a 'path' field.
        """
        if entry.is_dir():
            return self.render_directory(entry.path)

        name = self.render_name(entry)

        if entry.dir:
            dirname = self.render_directory(entry.dir + os.sep)
            path = String(dirname.value + name.value, dirname.length + name.length)
        else:
            path = name

        if entry.is_symlink() and with_link:
            return String(path.value + " -> " + entry.link, path.length + 4 + len(entry.link))
        else:
            return path

    def render_name(self, entry):
        """Render a 'name' field.
        """
        if entry.is_symlink() and not self.args.follow_symlinks:
            if entry.broken:
                return self.render_orphan(entry.name)
            else:
                return self.render_symlink(entry.name)

        elif entry.is_executable():
            return self.render_executable(entry.name)

        else:
            return self.render_file(entry.ext, entry.name)

    def render_file(self, ext, name):
        """Render the name part of a regular file.
        """
        # pylint:disable=unused-argument
        return String(name, len(name))

    def render_directory(self, name):
        """Render the directory part of a path name.
        """
        return String(name, len(name))

    def render_executable(self, name):
        """Render the name part of an executable regular file.
        """
        return String(name, len(name))

    def render_symlink(self, name):
        """Render the name part of a symbolic link.
        """
        return String(name, len(name))

    def render_orphan(self, name):
        """Render the name part of a missing symlink target.
        """
        return String(name, len(name))

    def pad_string(self, string, width):
        """Pad a string with spaces. string.length is the net length of string.value (without ansi
           escape sequences).
        """
        fill = max(0, abs(width) - string.length)
        if width < 0:
            return string.value + " " * fill
        else:
            return " " * fill + string.value

    def format_field(self, entry, field):
        """Format a field for output, i.e. render colors if required, pad with spaces, etc.
        """
        string = self.render_field(entry, field)

        if field.width is not None:
            return self.pad_string(string, int(field.width))
        else:
            return string.value

    def process(self, entry):
        """Print a separator delimited list of fields for an Entry object.
        """
        output = []
        for field in self.fields:
            try:
                output.append(self.format_field(entry, field))
            except KeyError:
                if self.args.all or field.modifier == "n":
                    output.append("")
                else:
                    return

        self.write_line(self.field_separator.join(output))

    def _write_line(self, line):
        """Print a line of output handling encoding errors.
        """
        try:
            print(line, end=self.args.newline)
        except UnicodeEncodeError:
            print(line.encode(self.encoding, "backslashreplace").decode(self.encoding),
                    end=self.args.newline)

        # If we don't flush stdout, printed results may be lost. That is because when the main
        # process exits, all multiprocessing processes vanish instantly without pending output
        # buffers being written.
        sys.stdout.flush()


class ColorConsole(Console):
    """Provide colorful console output using ansi escape sequences. The rules are are read from the
       LS_COLORS environment variable. For more info see man dir_colors(5).
    """

    def __init__(self, context):
        super().__init__(context)

        self.colors = {}
        self.extensions = {}
        rules = {}

        def esc(seq):
            return f"\x1b[{seq}m" if seq else ""

        colors = os.environ.get("LS_COLORS", "")
        if not colors:
            self.logger.warning("You chose console colour output "\
                    "but the LS_COLORS environment variable is not set")
            return

        # The content of LS_COLORS is a pathsep separated list of key=value pairs. File type rules
        # consist of two characters, extensions have the form '*.ext'.
        for part in colors.split(os.pathsep):
            try:
                key, value = part.split("=", 1)
            except ValueError:
                continue
            else:
                if len(key) == 2 and key.isalpha():
                    rules[key] = value
                else:
                    ext = os.path.splitext(key)[1][1:]
                    if ext:
                        self.extensions[ext] = esc(value)

        self.reset = esc(rules.get("rs", "0"))

        # Prepare escape sequences for directories, executables, symlinks and orphans. There are
        # some more we might add in the future.
        for rule in ("di", "ex", "ln", "or"):
            if rule in rules:
                self.colors[rule] = (esc(rules[rule]), self.reset)
            else:
                self.colors[rule] = None

    def render_file(self, ext, name):
        if ext in self.extensions:
            return String(f"{self.extensions[ext]}{name}{self.reset}", len(name))
        else:
            return String(name, len(name))

    def render_directory(self, name):
        return self.colorize("di", name)

    def render_executable(self, name):
        return self.colorize("ex", name)

    def render_symlink(self, name):
        return self.colorize("ln", name)

    def render_orphan(self, name):
        return self.colorize("or", name)

    def colorize(self, rule, text):
        """Return a String object with text colorized according to rule.
        """
        pair = self.colors[rule]
        if pair is None:
            return String(text, len(text))
        else:
            return String(f"{pair[0]}{text}{pair[1]}", len(text))
